package PScm::Env;

use strict;
use warnings;
use base qw(PScm::Expr);

use Carp;

sub new {
    my ($class, %bindings) = @_;

    $class = ref($class) || $class;
    bless { bindings => {%bindings}, parent => 0 }, $class;
}

sub Extend {
    my ($self, $symbols, $values) = @_;

    return $self->ExtendUnevaluated($symbols,
                                    $values->map_eval($self));
}

sub ExtendUnevaluated {
    my ($self, $symbols, $values) = @_;

    my %bindings;
    $self->_populate_bindings(\%bindings, $symbols, $values);
    my $newenv = $self->new(%bindings);
    $newenv->{parent} = $self;
    return $newenv;
}

sub _populate_bindings {
    my ($self, $bindings, $symbols, $values) = @_;
    if (ref($symbols) eq 'ARRAY') {
        confess "symbols is array";
    }
    if ($symbols->is_null) {
        die "too many arguments\n" unless $values->is_null;
    } elsif ($symbols->is_pair) {
        if ($values->is_pair) {
            my $symbol = $symbols->first;
            if ($symbol->is_symbol) {
                $bindings->{$symbol->value} = $values->first;
                $self->_populate_bindings($bindings,
                                          $symbols->rest,
                                          $values->rest);
            } else {
                die "bad formal arguments[1]:",
                    $symbol->as_string(), "\n";
            }
        } else {
            confess "not enough arguments, need: ", $symbols->as_string(), " got: ", $values->as_string(), "\n";
        }
    } elsif ($symbols->is_symbol) {
        $bindings->{$symbols->value} = $values;
    } else {
        die "bad formal arguments[2]:",
            $symbols->as_string(), "\n";
    }
}

sub ExtendRecursively {
    my ($self, $symbols, $values) = @_;

    my $newenv = $self->ExtendUnevaluated($symbols, $values);
    $newenv->_eval_values();
    return $newenv;
}

sub ExtendIteratively {
    my ($self, $symbols, $values) = @_;

    if ($symbols->is_null) {
        return $self;
    } else {
        return $self->Extend(new PScm::Expr::List($symbols->first),
                             new PScm::Expr::List($values->first))
                    ->ExtendIteratively($symbols->rest,
                                        $values->rest);
    }
}

sub _eval_values {
    my ($self) = @_;
    map {
        $self->{bindings}{$_} =
          $self->{bindings}{$_}->Eval($self)
      }
      keys %{ $self->{bindings} };
}

sub LookUp {
    my ($self, $symbol) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        return $$ref;
    } else {
        die "no binding for @{[$symbol->value]}",
          " in @{[ref($self)]}\n";
    }
}

sub Assign {
    my ($self, $symbol, $value) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        $$ref = $value;
    } else {
        die "no binding for @{[$symbol->value]}",
          " in @{[ref($self)]}\n";
    }
}

sub _lookup_ref {
    my ($self, $symbol) = @_;

    if (exists($self->{bindings}{ $symbol->value })) {
        return \$self->{bindings}{ $symbol->value };
    } elsif ($self->{parent}) {
        return $self->{parent}->_lookup_ref($symbol);
    } else {
        return undef;
    }
}

sub _lookup_method {
    my ($self, $method_name) = @_;

    return $self->_lookup_method_here($method_name)
      || $self->_lookup_method_in_super($method_name);
}

sub _lookup_method_here {
    my ($self, $method_name) = @_;

    if (exists $self->{bindings}{class}) {
        return $self->{bindings}{class}
          ->get_method($method_name, $self);
    }
}

sub _lookup_method_in_super {
    my ($self, $method_name) = @_;

    if (exists $self->{bindings}{super}) {
        return $self->{bindings}{super}
          ->_lookup_method($method_name);
    }
}

sub Define {
    my ($self, $symbol, $value) = @_;

    $self->{bindings}{ $symbol->value } = $value;
    return $symbol;
}

sub call_method {
    my ($self, $this, $method_name, $args, $env) = @_;

    if (my $method = $self->_lookup_method($method_name)) {
        return $method->ApplyMethod($this, $args, $env);
    }
}

sub Apply {
    my ($self, $form, $env) = @_;

    my ($method_symbol, $args) = ($form->first, $form->rest);
    my $res =
      $self->call_method($self, $method_symbol->value, $args, $env);

    if (defined $res) {
        return $res;
    } else {
        die "method ", $method_symbol->value, " not found\n";
    }
}

#########################
package PScm::Env::Super;

use base qw(PScm::Env);

sub Apply {
    my ($self, $form, $env) = @_;

    my ($method, $args) = ($form->first, $form->rest);
    my $this = $env->LookUp(PScm::Expr::Symbol->new("this"));
    my $res  =
      $self->call_method($this, $method->value, $args, $env);

    if (defined $res) {
        return $res;
    } else {
        die "method ", $method->value, " not found in super\n";
    }
}

1;
