package PScm::SpecialForm;

use strict;
use warnings;
use base qw(PScm::Expr);

###############################
package PScm::SpecialForm::Let;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env) = @_;

    my ($symbols, $values, $body) = $self->UnPack($form);

    return $body->Eval($env->Extend($symbols, $values));
}

sub UnPack {
    my ($self, $form) = @_;

    my ($bindings, $body) = $form->value;
    my ($symbols, $values) = $self->unpack_bindings($bindings);
    return ($symbols, $values, $body);
}

sub unpack_bindings {
    my ($self, $bindings) = @_;
    if ($bindings->is_null) {
        my $null = new PScm::Expr::List::Null();
        return ($null, $null);
    } else {
        my ($symbols, $values) =
            $self->unpack_bindings($bindings->rest);
        return (
            PScm::Expr::List->Cons($bindings->first->first,
                                   $symbols),
            PScm::Expr::List->Cons($bindings->first->rest->first,
                                   $values)
        );
    }
}

##################################
package PScm::SpecialForm::LetRec;

use base qw(PScm::SpecialForm::Let);

sub Apply {
    my ($self, $form, $env) = @_;

    my ($symbols, $values, $body) = $self->UnPack($form);

    return $body->Eval(
        $env->ExtendRecursively($symbols, $values));
}

###################################
package PScm::SpecialForm::LetStar;

use base qw(PScm::SpecialForm::Let);

sub Apply {
    my ($self, $form, $env) = @_;

    my ($symbols, $values, $body) = $self->UnPack($form);

    return $body->Eval(
        $env->ExtendIteratively($symbols, $values));
}

##############################
package PScm::SpecialForm::If;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env) = @_;

    my $condition = $form->first;
    my $true_branch = $form->rest->first;
    my $false_branch = $form->rest->rest->first;

    if ($condition->Eval($env)->isTrue) {
        return $true_branch->Eval($env);
    } else {
        return $false_branch->Eval($env);
    }
}

################################
package PScm::SpecialForm::Eval;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env) = @_;
    $form->first()->Eval($env)->Eval($env);
}

##################################
package PScm::SpecialForm::Lambda;

use base qw(PScm::SpecialForm);
use PScm::Closure;

sub Apply {
    my ($self, $form, $env) = @_;

    my ($args, $body) = $form->value;
    return PScm::Closure::Function->new($args, $body, $env);
}

##################################
package PScm::SpecialForm::Macro;

use base qw(PScm::SpecialForm);
use PScm::Closure;

sub Apply {
    my ($self, $form, $env) = @_;
    my ($args, $body) = $form->value;
    return PScm::Closure::Macro->new($args, $body, $env);
}

#################################
package PScm::SpecialForm::Quote;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env) = @_;
    return $form->first->Quote($env);
}

###############################
package PScm::SpecialForm::Set;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env) = @_;
    my ($symbol, $expr) = $form->value;
    $env->Assign($symbol, $expr->Eval($env));
}

#################################
package PScm::SpecialForm::Begin;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env) = @_;
    my (@expressions) = $form->value;

    my $ret = new PScm::Expr::List();

    while (@expressions) {
        $ret = shift(@expressions)->Eval($env);
    }

    return $ret;
}

##################################
package PScm::SpecialForm::Define;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env) = @_;
    my ($symbol, $expr) = $form->value;
    $env->Define($symbol, $expr->Eval($env));
}

#####################################
package PScm::SpecialForm::MakeClass;

use base qw(PScm::SpecialForm);

sub Apply {
    my ($self, $form, $env) = @_;

    my $parent_expr = $form->first;
    my $fields = $form->rest->first;
    my $methods = $form->rest->rest;
    my $parent_class = $parent_expr->Eval($env);
    return PScm::Class->new($parent_class,
                            $fields,
                            $methods,
                            $env);
}

1;
